(*
**
** untyped \lambda terms
**
*)


(*
** Definition of terms 
*)

(* mu[Tm:Set](var : Nat -> Tm,
	      app : Tm -> Tm -> Tm,
	      lam : Tm -> Tm)
*)

$[Tm : Set]
$[var : Nat -> Tm]
$[app : Tm -> Tm -> Tm]
$[lam : Tm -> Tm]

$[RecTm : {P:Tm->Type}
	   ({n:Nat}P (var n))
	-> ({M|Tm}(P M)->{N|Tm}(P N)->P (app M N))
	-> ({M|Tm}(P M)->P (lam M))
	-> {M:Tm}P M];

[[P:Tm->Type][var_:{n:Nat}P (var n)]
 [app_:{M|Tm}(P M)->{N|Tm}(P N)->P (app M N)]
 [lam_:{M|Tm}(P M)->P (lam M)]
 [n:Nat][M,N:Tm]
   RecTm P var_ app_ lam_ (var n) ==> var_ n
|| RecTm P var_ app_ lam_ (app M N) ==> app_ (RecTm P var_ app_ lam_ M)
					   (RecTm P var_ app_ lam_ N)
|| RecTm P var_ app_ lam_ (lam M) ==> lam_ (RecTm P var_ app_ lam_ M)];

[RecTmN [P|Type] = RecTm ([_:Tm]P)
 : (Nat->P)->(Tm->P->Tm->P->P)->(Tm->P->P)->Tm->P];

(*
** Simple lemmas about terms
*)

Goal {i:Nat}{M,M':Tm}not (EQ (app M M') (var i));
Intros i M M' H;
Refine EQ_rewrite H
         (RecTmN ([_:Nat]absurd) ([_:Tm][_:Set][_:Tm][_:Set]Unit)
	         ([_:Tm][_:Set]Unit));
Refine void;
Save neq_var_app;

Goal {i:Nat}{M:Tm}not (EQ (lam M)  (var i));
Intros i M H;
Refine EQ_rewrite H
         (RecTmN ([_:Nat]absurd) ([_:Tm][_:Set][_:Tm][_:Set]Unit)
	         ([_:Tm][_:Set]Unit));
Refine void;
Save neq_var_lam;

Goal {N,M1,M2:Tm}not (EQ (lam N) (app M1 M2));
Intros N M1 M2 H;
Refine EQ_rewrite H
         (RecTmN ([_:Nat]Unit) ([_:Tm][_:Set][_:Tm][_:Set]absurd)
	         ([_:Tm][_:Set]Unit));
Refine void;
Save neq_lam_app;

Goal  {M1,M2,N1,N2:Tm}(EQ (app M1 M2) (app N1 N2))
		->((EQ M1 N1)/\(EQ M2 N2));
intros;
Refine EQ_rewrite H 
         (RecTmN ([_:Nat]Unit) 
		 ([N1':Tm][_:Set][N2':Tm][_:Set](EQ M1 N1')/\(EQ M2 N2'))
	         ([_:Tm][_:Set]Unit));
andI;Refine EQ_refl;Refine EQ_refl;
Save inj_app;

Goal {M1,M2:Tm}(EQ (lam M1) (lam M2)) -> (EQ M1 M2);
intros;
Refine EQ_rewrite H 
         (RecTmN ([_:Nat]Unit) ([_:Tm][_:Set][_:Tm][_:Set]Unit)
	         ([M2':Tm][_:Set]EQ M1 M2'));
Refine EQ_refl;
Save inj_lam;

(* 
** weakening & substitution 
*)

(* weak_var : Nat -> Nat -> Nat 

rec weak_var zero     i = (succ i)
  | weak_var (succ l) zero = zero
  | weak_var (succ l) (succ i) = succ (weak_var l i)

*)

[weak_var =
  RecNatN succ
	  ([_:Nat][weak_l:Nat->Nat]
	   RecNatN zero ([i:Nat][_:Nat]succ (weak_l i)))];

(* weak : Nat->Tm->Tm

rec weak l (var i) = var (weak_var l i)
  | weak l (app M N) = app (weak l M) (weak l N)
  | weak l (lam M) = lam (weak (succ l) M)
*)

[weak[l:Nat][M:Tm] =
  RecTmN ([i:Nat][l:Nat]var (weak_var l i))
	 ([_:Tm][weak_M:Nat->Tm]
	  [_:Tm][weak_N:Nat->Tm]
	    [l:Nat]app (weak_M l) (weak_N l))
	 ([_:Tm][weak_M:Nat->Tm]
	    [l:Nat]lam (weak_M (succ l)))
	M l]
	
[weak0 = weak zero : Tm -> Tm];

[rep_weak0[M:Tm] = RecNatN M ([_:Nat][M:Tm]weak0 M)];

Goal {n:Nat}EQ (rep_weak0 (var zero) n) (var n);
Refine RecNat [n:Nat]EQ (rep_weak0 (var zero) n) (var n);
Refine EQ_refl;
intros;Equiv EQ ? (weak0 (var n));Refine EQ_resp weak0;Refine H;
Save rep_weak0_lem;

(* subst_var : Nat->Nat->Tm->Tm 

rec subst_var zero zero N = N
  | subst_var zero (succ i) N = var i
  | subst_var (succ l) zero N = var zero
  | subst_var (succ l) (succ i) N = weak0 (subst_var l i N)
*)

[subst_var =
  RecNatN (RecNatN ([N:Tm]N) ([i:Nat][_:Tm->Tm][_:Tm]var i))
	  ([l:Nat][subst_var_l : Nat->Tm->Tm]
	     RecNatN ([_:Tm]var zero)
		([i:Nat][_:Tm->Tm][N:Tm]weak0 (subst_var_l i N)))];

(* subst : Nat->Tm->Tm->Tm

rec subst l (var i) N = subst_var l i N
  | subst l (app M M') N = app (subst l M) (subst l N)
  | subst l (lam M) N = lam (subst (succ l) M N)

*)

[subst[l:Nat][M,N:Tm] =
  RecTmN ([i:Nat][l:Nat]subst_var l i N)
	 ([_:Tm][subst_M:Nat->Tm][_:Tm][subst_N:Nat->Tm]
		[l:Nat]app (subst_M l) (subst_N l))
	 ([_:Tm][subst_M:Nat->Tm]
		[l:Nat]lam (subst_M (succ l)))
	M l];


[subst0 = subst zero];

(*
** substitution & weakening laws 
*)

Goal {l:Nat}{M,N:Tm}EQ (subst l (weak l M) N) M;
intros;
Refine RecTm [M:Tm]{l:Nat}EQ (subst l (weak l M) N) M;
(* var *)
intros;
Refine RecNat [l:Nat]{i:Nat}EQ (subst l (weak l (var i)) N) (var i);
intros;Refine EQ_refl;
intros l2 IH i;
Refine RecNat [i:Nat]EQ (subst (succ l2) (weak (succ l2) (var i)) N) (var i);
Refine EQ_refl;
intros i1 _;
Equiv EQ (weak0 (subst l2 (weak l2 (var i1)) N)) (weak0 (var i1));
Refine EQ_resp weak0;Refine IH;
(* app *)
intros;Refine EQ_rewrite' (H l1) ([X:Tm]EQ (app X ?) ?);
Refine EQ_rewrite' (H1 l1) ([X:Tm]EQ (app ? X) ?);
Refine EQ_refl;
(* lam *)
intros;Refine EQ_rewrite' (H (succ l1)) ([X:Tm]EQ (lam X) ?);
Refine EQ_refl;
Save subst_weak_lem;

Goal {l',l:Nat}{M:Tm}(LE l' l)->
	(EQ (weak (succ l) (weak l' M))
	    (weak l' (weak l M)));
intros;
Refine RecTm [M:Tm]{l',l:Nat}(LE l' l)->
	 EQ (weak (succ l) (weak l' M)) (weak l' (weak l M));
(*** var ***)
intros i l1' l1 _;Refine EQ_resp var;
Refine RecLEN [l1',l1:Nat]{i:Nat}
	 EQ (weak_var (succ l1) (weak_var l1' i)) 
	    (weak_var l1' (weak_var l1 i));
(** LE0 **)
intros;Refine EQ_refl;
(** LE1 **)
intros l2' l2 _ IH;
Refine RecNat [i:Nat]
	EQ (weak_var (succ (succ l2)) (weak_var (succ l2') i)) 
	   (weak_var (succ l2') (weak_var (succ l2) i));
(* i=0 *) Refine EQ_refl;
(* succ i *)
intros i' _;Refine EQ_resp succ;Refine IH;
(** **) Immed;
(*** app ***)
intros;
Refine EQ_rewrite ? ([X:Tm]EQ ? (app X ?));Next +1;Refine H1;Immed;
Refine EQ_rewrite ? ([X:Tm]EQ ? (app ? X));Next +1;Refine H2;Immed;
Refine EQ_refl;
(*** lam ***)
intros;
Refine EQ_rewrite ? ([X:Tm]EQ ? (lam X));Next +1;Refine H1;Refine LE1;Immed;
Refine EQ_refl;
(*** ***) Immed;
Save weak_weak_lem;

Goal {l',l:Nat}{M,N:Tm}(LE l' l)->
	(EQ (subst (succ l) (weak l' M) N) 
	    (weak l' (subst l M N)));
intros;
Refine RecTm [M:Tm]{l',l:Nat}(LE l' l)->
	(EQ (subst (succ l) (weak l' M) N) 
	    (weak l' (subst l M N)));
(*** var ***)
intros i l1' l1 _;
Refine RecLEN [l',l:Nat]{i:Nat}
	EQ (subst (succ l) (weak l' (var i)) N) (weak l' (subst l (var i) N));
(** LE0 **)
intros;Refine EQ_refl;
(** LE_1 **)
intros l2' l2 _ IH;
Refine RecNat [i:Nat]
	EQ (subst (succ (succ l2)) (weak (succ l2') (var i)) N) 
	   (weak (succ l2') (subst (succ l2) (var i) N));
(* i = 0 *) Refine EQ_refl;
(* succ i *) 
intros i' _;
Refine EQ_trans;Next +2;Refine EQ_sym;Refine weak_weak_lem;Refine LE0;
Refine EQ_resp weak0;Refine IH;
(* *) Immed;
(*** app ***)
intros;
Refine EQ_rewrite ? ([X:Tm]EQ ? (app X ?));Next +1;Refine H1;Immed;
Refine EQ_rewrite ? ([X:Tm]EQ ? (app ? X));Next +1;Refine H2;Immed;
Refine EQ_refl;
(*** lam ***)
intros;
Refine EQ_rewrite ? ([X:Tm]EQ ? (lam X));Next +1;Refine H1;Refine LE1;Immed;
Refine EQ_refl;
(*** ***) Immed;
Save subst_weak_lem';

Goal {l:Nat}{M,N1,N2:Tm}
	EQ (subst l (subst (succ l) M N1) N2)
	   (subst l (subst l M (weak0 N2)) N1);
intros;
Refine RecTm [M:Tm]{l:Nat}
  EQ (subst l (subst (succ l) M N1) N2) (subst l (subst l M (weak0 N2)) N1);
(*** var ***)
intros;
Refine RecNat [l:Nat]{i:Nat}
   EQ (subst l (subst (succ l) (var i) N1) N2) 
      (subst l (subst l (var i) (weak0 N2)) N1);
(** l = 0 **)
Refine RecNat [i:Nat]
   EQ (subst zero (subst (succ zero) (var i) N1) N2) 
      (subst zero (subst zero (var i) (weak0 N2)) N1);
(* i = 0 *)
Refine EQ_trans;Next +2;Refine EQ_sym;Refine subst_weak_lem;Refine EQ_refl;
(* succ i *)
intros;Refine EQ_trans;Next +1;Refine subst_weak_lem;Refine EQ_refl;
(** succ l **)
intros l2 IH;
Refine RecNat [i:Nat]
   EQ (subst (succ l2) (subst (succ (succ l2)) (var i) N1) N2) 
      (subst (succ l2) (subst (succ l2) (var i) (weak0 N2)) N1);
(* i = 0 *)
Refine EQ_refl;
(* succ i *)
intros i _;
Refine EQ_trans;Next +1;Refine subst_weak_lem';Refine LE0;
Refine EQ_trans;Next +2;Refine EQ_sym;Refine subst_weak_lem';Refine LE0;
Refine EQ_resp weak0;Refine IH;
(* app *)
intros;Refine EQ_rewrite (H l1) ([X:Tm]EQ ? (app X ?));
Refine  EQ_rewrite (H1 l1) ([X:Tm]EQ ? (app ? X));Refine EQ_refl;
(* lam *)
intros;Refine EQ_rewrite (H (succ l1)) ([X:Tm]EQ ? (lam X));
Refine EQ_refl;
Save subst_subst_lem;

Goal {m,l:Nat}{M1,M2,N:Tm}
	EQ (subst m (subst (succ (add m l)) M1 N) (subst l M2 N))
	   (subst (add m l) (subst m M1 M2) N);
intros m' l' M1' M2' N';
Refine RecTm [M1:Tm]{m,l:Nat}{M2,N:Tm}
	EQ (subst m (subst (succ (add m l)) M1 N) (subst l M2 N))
	   (subst (add m l) (subst m M1 M2) N);
(*** var ***)
intros;
Refine RecNat [i:Nat]{m:Nat}
	EQ (subst m (subst (succ (add m l)) (var i) N) (subst l M2 N)) 
	   (subst (add m l) (subst m (var i) M2) N);
(** i=0 **)
Refine RecNat [m:Nat]
	EQ (subst m (var zero) (subst l M2 N)) 
	   (subst (add m l) (subst m (var zero) M2) N);
(* m=0 *) Refine EQ_refl;
(* succ m *) intros; Refine EQ_refl;
(** succ i **)
intros i IH;
Refine RecNat [m:Nat]
	EQ (subst m (subst (succ (add m l)) (var (succ i)) N) (subst l M2 N)) 
	   (subst (add m l) (subst m (var (succ i)) M2) N);
(* m=0 *) intros; Refine subst_weak_lem;
(* succ m *) 
intros i1 _;
Refine EQ_trans;Next +1;Refine subst_weak_lem';Refine LE0;
Refine EQ_trans;Next +2;Refine EQ_sym;Refine subst_weak_lem';Refine LE0;
Refine EQ_resp weak0;Refine IH;
(*** app ***)
intros;
Refine EQ_rewrite ? ([X:Tm]EQ ? (app X ?));Next +1;Refine H;
Refine EQ_rewrite ? ([X:Tm]EQ ? (app ? X));Next +1;Refine H1;
Refine EQ_refl;
(*** lam ***)
intros;
Refine EQ_rewrite ? ([X:Tm]EQ ? (lam X));Next +1;Refine H;
Refine EQ_refl;
Save subst_subst_lem';

(*
** one-step reduction
*)

(* mu[Step:Tm->Tm->Set](
	beta : {M,N:Tm}Step (app (lam M) N) (subst0 M N),
	app_l : {M,M',N:Tm}(Step M M')->(Step (app M N) (app M' N)),
	app_r : {M,M',N:Tm}(Step M M')->(Step (app N M) (app N M')),
	xi : {M,N:Tm}(Step M M')->(Step (lam M) (lam M')) )
*)

$[Step:Tm->Tm->Set]
$[beta : {M,N:Tm}Step (app (lam M) N) (subst0 M N)]
$[app_l : {M,M',N:Tm}(Step M M')->(Step (app M N) (app M' N))]
$[app_r : {M,M',N:Tm}(Step M M')->(Step (app N M) (app N M'))]
$[xi : {M,M':Tm}(Step M M')->(Step (lam M) (lam M'))]
$[RecStep:{P:{M,N|Tm}(Step M N)->Type}
	({M,N:Tm}P (beta M N))
	-> ({M,M',N:Tm}{p:Step M M'}(P p)->(P (app_l M M' N p)))
	-> ({M,M',N:Tm}{p:Step M M'}(P p)->P (app_r M M' N p))
	-> ({M,M':Tm}{p:Step M M'}(P p)->P (xi M M' p))
	-> {M,N|Tm}{p:Step M N}P p];

[[P:{M,N|Tm}(Step M N)->Type]
 [beta_ : {M,N:Tm}P (beta M N)]
 [app_l_ : {M,M',N:Tm}{p:Step M M'}(P p)->P (app_l M M' N p)]
 [app_r_ : {M,M',N:Tm}{p:Step M M'}(P p)->P (app_r M M' N p)]
 [xi_ : {M,M':Tm}{p:Step M M'}(P p)->P (xi M M' p)]
 [M,M',N:Tm][p:Step M M']
   RecStep P beta_ app_l_ app_r_ xi_ (beta M N) ==> beta_ M N
|| RecStep P beta_ app_l_ app_r_ xi_ (app_l M M' N p)
	==> app_l_ M M' N p (RecStep P beta_ app_l_ app_r_ xi_ p)
|| RecStep P beta_ app_l_ app_r_ xi_ (app_r M M' N p)
	==> app_r_ M M' N p (RecStep P beta_ app_l_ app_r_ xi_ p)
|| RecStep P beta_ app_l_ app_r_ xi_ (xi M M' p)
	==> xi_ M M' p (RecStep P beta_ app_l_ app_r_ xi_ p)];

[RecStepN[P:Tm->Tm->Type] = RecStep ([M,N|Tm][_:Step M N]P M N)
 : ({M,N:Tm}P (app (lam M) N) (subst0 M N))
   ->({M,M',N:Tm}(Step M M')->(P M M')->P (app M N) (app M' N))
   ->({M,M',N:Tm}(Step M M')->(P M M')->P (app N M) (app N M'))
   ->({M,M':Tm}(Step M M')->(P M M')->P (lam M) (lam M'))
   ->{M,N|Tm}(Step M N)->P M N];

(*
** Lemmas about Step
*)

Goal {M,lam_M':Tm}(Step (lam M) lam_M')
	->Ex[M':Tm](Step M M') /\ (EQ lam_M' (lam M'));
intros;
Claim {lam_M,lam_M':Tm}(EQ (lam M) lam_M)->(Step (lam M) lam_M')
	->Ex ([M':Tm]and (Step M M') (EQ lam_M' (lam M')));
Refine ?+1;Next +1;Refine EQ_refl;Immed;
intros;
Refine RecStepN [lam_M,lam_M':Tm](EQ (lam M) lam_M)
	->Ex ([M':Tm]and (Step M M') (EQ lam_M' (lam M')));
intros;Refine neq_lam_app;Next +3;Refine H3;
intros;Refine neq_lam_app;Next +3;Refine H5;
intros;Refine neq_lam_app;Next +3;Refine H5;
intros;Refine ExIntro;Next +1;andI;
Refine EQ_rewrite' ? ([X:Tm]Step X ?);Next +1;Refine inj_lam;Refine H5;
Refine H3;Refine EQ_refl;
Next +1;Refine H2;Refine EQ_refl;
Save step_lam;

Goal {l:Nat}{M,M',N:Tm}(Step M M')->(Step (subst l M N) (subst l M' N));
intros;
Refine RecStepN [M,M':Tm]{l:Nat}Step (subst l M N) (subst l M' N);
(* beta *)
intros;
Refine EQ_rewrite ? [X:Tm]Step ? X;Next +1;Refine subst_subst_lem';
Refine beta;
(* app_l *)
intros;Refine app_l;Refine H2;
(* app_r *)
intros;Refine app_r;Refine H2;
(* xi *)
intros;Refine xi;Refine H2;
(* End of cases *)
Immed;
Save step_subst;

[nf[M:Tm] = {M':Tm}not (Step M M')];

Goal {i:Nat}(nf (var i));
Intros i M H;
Claim {M,M'|Tm}(Step M M')-> not (EQ M (var i));
Refine ?+1;Next +3;Refine EQ_refl;Immed;
Refine RecStepN [M,M':Tm]not (EQ M (var i));
intros;Refine neq_var_app;
intros;Refine neq_var_app;
intros;Refine neq_var_app;
intros;Refine neq_var_lam;
Save nf_var;

(* examples *)

(*
[T0 = var zero][T1 = lam T0];
[T2 = lam (app (var zero) (var (succ zero)))];
weak0 T0; Normal VReg;
weak0 T1; Normal VReg;
weak0 T2; Normal VReg;
subst0 T2 T1; Normal VReg;
subst0 T2 T0; Normal VReg;
*)





